package ch.epfl.ppo.collections;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.BitSet;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Random;

import org.junit.Test;

public abstract class SetTest {
    abstract <T extends Comparable<T>> Set<T> newSet();

    @Test
    public void isInitiallyEmpty() {
        assertTrue(newSet().isEmpty());
    }

    @Test
    public void sizeInitially0() {
        assertEquals(0, newSet().size());
    }

    @Test
    public void addAbsentIncreasesSize() {
        Set<String> s = newSet();
        String str = "a";
        for (int i = 0; i < 100; ++i) {
            assertEquals(i, s.size());
            s.add(str);
            str = str + "a";
        }
    }

    @Test
    public void addPresentDoesNotIncreaseSize() {
        Set<String> s = newSet();
        String[] strs = new String[] { "a", "b", "c", "d" };
        for (String str: strs)
            s.add(str);
        Random rng = new Random();
        for (int i = 0; i < 100; ++i) {
            assertEquals(strs.length, s.size());
            s.add(strs[rng.nextInt(strs.length)]);
        }
    }

    @Test
    public void removePresentDecreasesSize() {
        Set<Integer> s = newSet();
        for (int i = 0; i < 100; ++i)
            s.add(i);
        for (int i = 99; i >= 0; --i) {
            assertEquals(i + 1, s.size());
            s.remove(i);
        }
        assertTrue(s.isEmpty());
    }

    @Test
    public void removeAbsentDoesNotDecreaseSize() {
        Set<Integer> s = newSet();
        for (int i = 0; i < 100; ++i)
            s.add(-(i + 1));
        Random rng = new Random();
        for (int i = 0; i < 100; ++i) {
            assertEquals(100, s.size());
            s.remove(rng.nextInt(100));
        }
    }

    @Test
    public void removeRemovesSingleElement() {
        Set<Integer> s = newSet();
        s.add(12);
        s.remove(12);
        for (int i: s)
            fail("unexpected element: " + i);
    }

    @Test
    public void removeRemovesOneOutOfThree() {
        Set<Integer> s = newSet();
        s.add(10);
        s.add(9);
        s.add(11);
        s.remove(10);
        int c = 0;
        for (int i: s) {
            assertTrue(i == 9 || i == 11);
            ++c;
        }
        assertEquals(2, c);
    }

    @Test
    public void containsAbsentIsFalse() {
        Set<Integer> s = newSet();
        for (int i = 0; i < 100; ++i)
            assertFalse(s.contains(i));
        Random rng = new Random();
        for (int i = 0; i < 100; ++i)
            s.add(-(rng.nextInt(100) + 1));
        for (int i = 0; i < 100; ++i)
            assertFalse(s.contains(i));
    }

    @Test
    public void containsPresentIsTrue() {
        String[] strs = new String[] { "a", "b", "c", "ab", "ac" };
        Set<String> s = newSet();
        Random rng = new Random();
        int offset = rng.nextInt(strs.length);
        for (int i = 0; i < strs.length; ++i)
            s.add(strs[(offset + i) % strs.length]);
        for (int i = 0; i < strs.length; ++i)
            assertTrue(s.contains(strs[rng.nextInt(strs.length)]));
    }

    @Test
    public void iteratorIterates() {
        Set<Integer> s = newSet();
        for (int i = 0; i < 100; ++i)
            s.add(i);
        BitSet bs = new BitSet();
        bs.set(0, 100);
        for (int i: s) {
            assertTrue(bs.get(i));
            bs.flip(i);
        }
        assertEquals(0, bs.cardinality());
    }

    @Test
    public void iteratorRemoves() {
        Set<Integer> s = newSet();
        int[] primes = new int[] { 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97 };
        Set<Integer> primesSet = newSet();
        for (int p: primes)
            primesSet.add(p);
        for (int i = 0; i < 100; ++i)
            s.add(i);
        Iterator<Integer> it = s.iterator();
        while (it.hasNext()) {
            if (! primesSet.contains(it.next()))
                it.remove();
        }
        assertEquals(primes.length, s.size());
        for (int p: primes)
            assertTrue(s.contains(p));
    }

    @Test(expected = NoSuchElementException.class)
    public void iteratorNextFailsIfNotHasNext() {
        this.<Integer>newSet().iterator().next();
    }

    @Test(expected = IllegalStateException.class)
    public void iteratorRemoveFailsBeforeFirstNext() {
        this.<Integer>newSet().iterator().remove();
    }

    @Test(expected = IllegalStateException.class)
    public void iteratorRemoveFailsWhenCalledTwice() {
        Set<Integer> l = newSet();
        l.add(1);
        l.add(2);
        Iterator<Integer> it = l.iterator();
        it.next();
        it.remove();
        it.remove();
    }
}
